import typing as ta


T = ta.TypeVar('T')

Comparer: ta.TypeAlias = ta.Callable[[T, T], int]


##


class TreapNode(ta.Protocol[T]):
    @property
    def value(self) -> T: ...

    @property
    def priority(self) -> int: ...

    @property
    def left(self) -> ta.Optional['TreapNode[T]']: ...

    @property
    def right(self) -> ta.Optional['TreapNode[T]']: ...

    @property
    def count(self) -> int: ...

    def __iter__(self) -> ta.Iterator[T]: ...


##


from ._treap_py import (  # noqa
    new,
    find,
    place,
    union,
    split,
    intersect,
    delete,
    diff,
)


try:
    from . import _treap  # type: ignore
except ImportError:
    pass
else:
    globals().update({a: getattr(_treap, a) for a in [
        'new',
        'find',
        'place',
        'union',
        'split',
        'intersect',
        'delete',
        'diff',
    ]})



import dataclasses as dc
import typing as ta

from ..api.options import Option


##


@dc.dataclass(frozen=True, kw_only=True)
class DefaultIterableConstructors(Option):
    iterable: ta.Callable[[ta.Iterable], ta.Any] | ta.Any | None = None
    sequence: ta.Callable[[ta.Iterable], ta.Any] | ta.Any | None = None
    mutable_sequence: ta.Callable[[ta.Iterable], ta.Any] | ta.Any | None = None


@dc.dataclass(frozen=True, kw_only=True)
class DefaultMappingConstructors(Option):
    mapping: ta.Callable[[ta.Iterable], ta.Any] | ta.Any | None = None
    mutable_mapping: ta.Callable[[ta.Iterable], ta.Any] | ta.Any | None = None



# ruff: noqa: SLF001
import typing as ta

from .. import check
from .codecs import Codec
from .mappers import Mapper


##


class Registry:
    def __init__(
            self,
            mappers: ta.Iterable[Mapper],
            *,
            codec: Codec | None = None,
    ) -> None:
        super().__init__()

        self._codec = codec

        self._mappers: set[Mapper] = set()

        self._mappers_by_cls: dict[type, Mapper] = {}
        self._mappers_by_cls_name: dict[str, Mapper] = {}
        self._mappers_by_store_name: dict[str, Mapper] = {}

        self.register(*mappers)

    @property
    def mappers(self) -> ta.AbstractSet[Mapper]:
        return self._mappers

    @property
    def mappers_by_cls(self) -> ta.Mapping[type, Mapper]:
        return self._mappers_by_cls

    @property
    def mappers_by_cls_name(self) -> ta.Mapping[str, Mapper]:
        return self._mappers_by_cls_name

    @property
    def mappers_by_store_name(self) -> ta.Mapping[str, Mapper]:
        return self._mappers_by_store_name

    @property
    def codec(self) -> Codec | None:
        return self._codec

    #

    def _check_can_register(self, m: Mapper) -> None:
        check.not_in(m, self._mappers)
        check.not_in(m.cls, self._mappers_by_cls)
        check.not_in(m.cls.__name__, self._mappers_by_cls_name)
        check.not_in(m.store_name, self._mappers_by_store_name)

    def _register(self, m: Mapper) -> None:
        self._check_can_register(m)

        m._set_registry(self)

        self._mappers.add(m)
        self._mappers_by_cls[m.cls] = m
        self._mappers_by_cls_name[m.cls.__name__] = m
        self._mappers_by_store_name[m.store_name] = m

    def register(self, *mappers: Mapper) -> None:
        for m in mappers:
            self._register(m)

    #

    def mapper_for_cls(self, cls: type) -> Mapper:
        return self._mappers_by_cls[cls]

    def mapper_for_obj(self, obj: ta.Any) -> Mapper:
        return self.mapper_for_cls(type(obj))



import abc
import functools
import typing as ta

from .. import check
from .. import lang


K = ta.TypeVar('K')


##


@functools.total_ordering
class Key(lang.Sealed, lang.Abstract, ta.Generic[K]):
    @property
    @abc.abstractmethod
    def k(self) -> K:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def __lt__(self, other: object) -> bool:
        raise NotImplementedError


##


@ta.final
class _Key(Key[K], lang.Final):
    def __init__(self, k: K) -> None:
        check.not_in(k.__class__, _KEY_TYPES)
        self._k = k

    def __repr__(self) -> str:
        return f'orm.key({self._k!r})'

    @property
    def k(self) -> K:
        return self._k

    #

    def __hash__(self) -> int:
        return hash(self._k)

    def __eq__(self, other: object) -> bool:
        return other.__class__ is _Key and self._k == other._k  # type: ignore[attr-defined]

    def __lt__(self, other: object) -> bool:
        if (oc := other.__class__) is _Key:
            return self._k < other._k  # type: ignore[attr-defined]
        elif oc is _AutoKey:
            return False
        else:
            raise TypeError(other)


@ta.overload
def key(k: Key[K]) -> Key[K]:
    ...


@ta.overload
def key(k: K) -> Key[K]:
    ...


def key(k):
    if k.__class__ in _KEY_TYPES:
        return k
    return _Key(k)


##


class AutoKeyNotSetError(Exception):
    pass


@ta.final
class _AutoKey(Key[K], lang.Final):
    def __repr__(self) -> str:
        return f'orm.auto_key@{id(self):x}()'

    @property
    def k(self) -> K:
        raise AutoKeyNotSetError

    #

    def __lt__(self, other: object) -> bool:
        if (oc := other.__class__) is _Key:
            return True
        elif oc is _AutoKey:
            return id(self) < id(other)
        else:
            raise TypeError(other)


def auto_key() -> Key:
    return _AutoKey()


def is_auto_key(k: ta.Any) -> bool:
    return k.__class__ is _AutoKey


##


def _unwrap_key(k: Key) -> ta.Any:
    if k.__class__ is _AutoKey:
        return k
    return k.k


_KEY_TYPES: tuple[type[Key], ...] = (
    _Key,
    _AutoKey,
)



# ruff: noqa: SLF001
import contextvars
import typing as ta

from .. import check
from .. import lang
from .keys import _KEY_TYPES
from .keys import Key
from .keys import _AutoKey
from .keys import _Key
from .keys import _unwrap_key
from .mappers import Mapper
from .queries import Query
from .refs import _DirectRef
from .refs import _LazyRef
from .registries import Registry
from .snaps import Snap
from .stores import Store


K = ta.TypeVar('K')
T = ta.TypeVar('T')


##


class _NOT_SET(lang.Marker):  # noqa
    pass


class Session:
    def __init__(
            self,
            registry: Registry,
            store: Store,
            *,
            no_auto_flush: bool = True,
    ) -> None:
        super().__init__()

        self._registry = registry
        self._store = store
        self._no_auto_flush = no_auto_flush

        self._aborted = False

        self._entities_by_key_by_cls: dict[type, dict[Key, Session._Entity]] = {m._cls: {} for m in registry._mappers}
        self._entities_by_obj_id: dict[int, Session._Entity] = {}
        self._entities_by_auto_key: dict[_AutoKey, Session._Entity] = {}

    def __repr__(self) -> str:
        return f'{type(self).__name__}@{id(self):x}({"aborted=True" if self._aborted else ""})'

    #

    @property
    def aborted(self) -> bool:
        return self._aborted

    def abort(self) -> None:
        if self._aborted:
            return

        self._aborted = True

        del self._entities_by_key_by_cls
        del self._entities_by_obj_id
        del self._entities_by_auto_key

    #

    def activate(self) -> ta.ContextManager[ta.Self]:
        return _SessionActivator(self)

    #

    class _Entity:
        def __init__(
                self,
                m: Mapper,
                k: Key,
                *,
                obj: ta.Any | None = None,
                snap: Snap | None = None,
        ) -> None:
            super().__init__()

            self.m = m
            self.k = k

            self.obj = obj
            self.snap = snap

        m: ta.Final[Mapper]
        k: ta.Final[Key]

        def __repr__(self) -> str:
            return (
                f'{self.__class__.__name__}('
                f'{self.m!r}'
                f', {self.k!r}'
                f'{f", obj={self.obj!r}" if self.obj is not None else ""}'
                f'{f", snap={self.snap!r}" if self.snap is not None else ""}'
                f')'
            )

    def _attach(
            self,
            cls: type | None = None,
            k: Key | None = None,
            *,
            obj: ta.Any | type[_NOT_SET] | None = _NOT_SET,
            snap: Snap | type[_NOT_SET] | None = _NOT_SET,
    ) -> _Entity:
        check.state(not self._aborted)

        check.arg(not (
            obj is not _NOT_SET and
            snap is not _NOT_SET and
            (
                (obj is None and snap is not None) or
                (obj is not None and snap is None)
            )
        ))

        #

        if cls is None:
            check.is_not(obj, _NOT_SET)
            cls = type(check.not_none(obj))
        elif obj is not _NOT_SET and obj is not None:
            check.is_(type(obj), cls)

        m = self._registry.mapper_for_cls(cls)

        #

        if obj is not _NOT_SET and obj is not None:
            ok = m.key_for_obj(obj)
            if k is None:
                k = ok
            else:
                check.equal(k, ok)

        if snap is not _NOT_SET and snap is not None:
            sk = m.key_for_snap(snap)  # type: ignore[arg-type]
            if k is None:
                k = sk
            else:
                check.equal(k, sk)

        if obj is not _NOT_SET and obj is not None and snap is not _NOT_SET and snap is not None:
            check.equal(m.obj_to_snap(obj), snap)

        k = check.not_none(k)

        #

        ed = self._entities_by_key_by_cls[cls]

        try:
            e = ed[k]
        except KeyError:
            pass

        else:
            check.equal(e.k, k)

            if obj is not _NOT_SET:
                check.is_(e.obj, obj)
                if obj is not None:
                    check.is_(self._entities_by_obj_id[id(obj)], e)

            if snap is not _NOT_SET:
                check.equal(e.snap, snap)

            return e

        #

        if snap is not _NOT_SET and snap is not None and obj is _NOT_SET:
            obj = m.snap_to_obj(snap)  # type: ignore[arg-type]

        if obj is _NOT_SET:
            obj = None
        if snap is _NOT_SET:
            snap = None

        e = self._Entity(
            m,
            k,
            obj=obj,
            snap=snap,  # type: ignore[arg-type]
        )

        self._entities_by_key_by_cls.setdefault(m.cls, {})[k] = e

        if obj is not None:
            self._entities_by_obj_id[id(obj)] = e

        if k.__class__ is _AutoKey:
            self._entities_by_auto_key[k] = e  # type: ignore[index]  # noqa

        return e

    #

    def add(self, *objs: ta.Any) -> None:
        for obj in objs:
            self._attach(obj=obj)

    #

    @ta.overload
    def get(self, cls: type[T], k: Key[K]) -> T | None:
        ...

    @ta.overload
    def get(self, cls: type[T], k: K) -> T | None:
        ...

    def get(self, cls, k):  # noqa
        check.state(not self._aborted)

        if k.__class__ not in _KEY_TYPES:
            k = _Key(k)

        cd = self._entities_by_key_by_cls.setdefault(cls, {})

        try:
            e = cd[k]
        except KeyError:
            pass
        else:
            return e.obj

        m = self._registry.mapper_for_cls(cls)

        snap = self._store.fetch(m, k.k)
        if snap is None:
            return None

        e = self._attach(
            cls,
            snap=snap,
        )

        return e.obj

    #

    def _delete(self, obj: ta.Any) -> None:
        e = self._attach(obj=obj)

        e.obj = None

        del self._entities_by_obj_id[id(obj)]

    def delete(self, *objs: ta.Any) -> None:
        check.state(not self._aborted)

        for obj in objs:
            self._delete(obj=obj)

    #

    def flush(self) -> None:
        check.state(not self._aborted)

        entity_ops: list[tuple[
            Session._Entity,
            ta.Any | None,
            Snap | None,
        ]] = []

        for cls, cd in self._entities_by_key_by_cls.items():
            m = self._registry.mapper_for_cls(cls)

            insert: list[Snap] = []
            update: list[tuple[ta.Any, Snap]] = []
            delete: list[ta.Any] = []

            for e in cd.values():
                if e.obj is None:
                    if e.snap is None:
                        continue
                    delete.append(_unwrap_key(e.k))
                    entity_ops.append((e, None, None))

                else:
                    snap = e.m.obj_to_snap(e.obj)
                    if snap == e.snap:
                        continue
                    entity_ops.append((e, e.obj, snap))
                    if (xs := e.snap) is None:
                        insert.append(snap)
                    else:
                        ds = {k: nv for k, nv in snap.items() if xs[k] != nv}
                        update.append((_unwrap_key(e.k), ds))

            if not (insert or update or delete):
                continue

            batch = Store.FlushBatch(
                insert=insert or None,
                update=update or None,
                delete=delete or None,
            )

            self._store.flush(m, batch)

        for e, e_obj, e_snap in entity_ops:
            if e_obj is None and (xo := e.obj) is not None:
                del self._entities_by_obj_id[id(xo)]
            e.obj = e_obj
            e.snap = e_snap

    #

    def _get_direct_ref_key(self, dr: _DirectRef) -> Key:
        e = self._entities_by_obj_id[id(dr._obj)]
        return e.k

    def _get_lazy_ref_obj(self, lr: _LazyRef) -> ta.Any:
        # TODO: writeback?
        return check.not_none(self.get(lr._cls, lr._k))

    #

    def query(self, q: Query[T]) -> list[T]:
        check.state(not self._aborted)

        if not self._no_auto_flush:
            self.flush()

        cls = q._cls
        m = self._registry.mapper_for_cls(cls)

        wh: dict[str, ta.Any] = {}
        for k, v in q.where.items():
            f = m._fields_by_name[k]
            wh[f._store_name] = m.field_value_to_snap_value(f, v)

        if not (snaps := self._store.lookup(m, wh)):
            return []

        out: list[T] = []

        for snap in snaps:
            e = self._attach(cls, snap=snap)

            if (obj := e.obj) is not None:
                out.append(obj)

        return out


##


class NoActiveSessionError(Exception):
    pass


class SessionAlreadyActiveError(Exception):
    pass


_ACTIVE_SESSION: contextvars.ContextVar[Session] = contextvars.ContextVar(f'{__name__}._ACTIVE_SESSION')


@ta.final
class _SessionActivator:
    def __init__(self, s) -> None:
        self._s = s

    _tok: ta.Any

    def __enter__(self):
        try:
            x = _ACTIVE_SESSION.get()
        except LookupError:
            pass
        else:
            raise SessionAlreadyActiveError(x)

        self._tok = _ACTIVE_SESSION.set(self._s)
        return self._s

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            if not self._s._aborted and not self._s._no_auto_flush:
                self._s.flush()
        else:
            self._s.abort()

        _ACTIVE_SESSION.reset(self._tok)


def active_session() -> Session:
    try:
        return _ACTIVE_SESSION.get()
    except LookupError:
        raise NoActiveSessionError from None


def opt_active_session() -> Session | None:
    try:
        return _ACTIVE_SESSION.get()
    except LookupError:
        return None



import typing as ta

from .. import check
from .. import lang
from .. import reflect as rfl
from .keys import Key
from .refs import Ref


##


class Field(lang.Sealed):
    def __init__(
            self,
            *,
            _name: str,
            _store_name: str,
            _rty: rfl.Type,
    ) -> None:
        super().__init__()

        self._name = check.non_empty_str(_name)
        self._store_name = check.non_empty_str(_store_name)
        self._rty = check.isinstance(_rty, rfl.Type)

    @classmethod
    def _default_store_name(cls, name: str) -> str:
        return name

    def __repr__(self) -> str:
        return ''.join([
            f'{type(self).__name__}(',
            ', '.join([
                repr(self._name),
                repr(self._rty),
            ]),
            ')',
        ])

    @property
    def name(self) -> str:
        return self._name

    @property
    def store_name(self) -> str:
        return self._store_name

    @property
    def rty(self) -> rfl.Type:
        return self._rty


@ta.final
class KeyField(Field, lang.Final):
    def __init__(self, **kwargs: ta.Any) -> None:
        super().__init__(**kwargs)

        gty = check.isinstance(self._rty, rfl.Generic)
        check.is_(gty.cls, Key)
        self._key_cls = check.isinstance(check.single(gty.args), type)

    @property
    def key_cls(self) -> type:
        return self._key_cls


@ta.final
class RefField(Field, lang.Final):
    def __init__(self, **kwargs: ta.Any) -> None:
        super().__init__(**kwargs)

        if isinstance(rty := self._rty, rfl.Union):
            check.state(rty.is_optional)
            rty = rty.without_none()
            self._optional = True
        else:
            self._optional = False

        gty = check.isinstance(rty, rfl.Generic)
        check.is_(gty.cls, Ref)
        ort, krt = gty.args
        self._ref_obj_cls = check.isinstance(ort, type)
        self._ref_key_cls = check.isinstance(krt, type)

    @property
    def optional(self) -> bool:
        return self._optional

    @property
    def ref_obj_cls(self) -> type:
        return self._ref_obj_cls

    @property
    def ref_key_cls(self) -> type:
        return self._ref_key_cls



from .api import (  # noqa
    index,
    field,
    mapper,
    dataclass_mapper,
    registry,

    session,
    abort,
    add,
    get,
    delete,
    flush,

    make_query,
    query,
)

from .codecs import (  # noqa
    Codec,
    NopCodec,
    FnCodec,

    MarshalCodec,
)

from .fields import (  # noqa
    Field,
    KeyField,
    RefField,

)

from .indexes import (  # noqa
    Index,
)

from .keys import (  # noqa
    Key,

    key,

    AutoKeyNotSetError,
    auto_key,
    is_auto_key,
)

from .inmemory import (  # noqa
    InMemoryStore,
)

from .mappers import (  # noqa
    Mapper,
)

from .queries import (  # noqa
    Query,
)

from .refs import (  # noqa
    Ref,

    ref,
)

from .registries import (  # noqa
    Registry,
)

from .sessions import (  # noqa
    Session,

    NoActiveSessionError,
    SessionAlreadyActiveError,
    active_session,
    opt_active_session,
)

from .stores import (  # noqa
    Store,
)

from .wrappers import (  # noqa
    WRAPPER_TYPES,
)



"""
https://github.com/python/typing/issues/213
"""
import typing as ta

from .. import lang
from .persistent import PersistentMapping
from .sorted import SortedMapping


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


class PersistentSortedMapping(
    PersistentMapping[K, V],
    SortedMapping[K, V],
    lang.Abstract,
    ta.Generic[K, V],
):
    pass



# ruff: noqa: SLF001
import abc
import typing as ta

from .. import check
from .. import lang
from .keys import _KEY_TYPES
from .keys import Key
from .keys import key


if ta.TYPE_CHECKING:
    from . import sessions as _sessions
    from . import wrappers as _wrappers
else:
    _sessions = lang.proxy_import('.sessions', __package__)
    _wrappers = lang.proxy_import('.wrappers', __package__)


K = ta.TypeVar('K')
T = ta.TypeVar('T')


##


class Ref(lang.Sealed, lang.Abstract, ta.Generic[T, K]):
    @property
    @abc.abstractmethod
    def cls(self) -> type[T]:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def k(self) -> Key[K]:
        raise NotImplementedError

    @abc.abstractmethod
    def __call__(self) -> T:
        raise NotImplementedError

    _hash: int

    @ta.final
    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            pass
        self._hash = h = hash((self.cls, self.k))
        return h

    @ta.final
    def __eq__(self, other):
        if self is other:
            return True
        ot = other.__class__
        if ot not in _REF_TYPES:
            return False
        return ot.cls == self.cls and ot.k == self.k


#


@ta.final
class _DirectRef(Ref[T, K], lang.Final):
    def __init__(self, obj: T) -> None:  # noqa
        cls = type(obj)
        check.not_in(cls, _wrappers.WRAPPER_TYPES)
        self._cls, self._obj = cls, obj

    def __repr__(self) -> str:
        return f'orm.ref({self._obj!r})'

    @property
    def cls(self) -> type[T]:
        return self._cls

    @property
    def k(self) -> Key[K]:
        return _sessions.active_session()._get_direct_ref_key(self)

    def __call__(self) -> T:
        return self._obj


#


@ta.final
class _LazyRef(Ref[T, K], lang.Final):
    def __init__(self, cls: type[T], k: Key[K]) -> None:  # noqa
        check.in_(k.__class__, _KEY_TYPES)
        self._cls, self._k = cls, k

    def __repr__(self) -> str:
        return f'orm.ref({self._cls.__name__}, {self._k!r})'

    @property
    def cls(self) -> type[T]:
        return self._cls

    @property
    def k(self) -> Key[K]:
        return self._k

    def __call__(self) -> T:
        return _sessions.active_session()._get_lazy_ref_obj(self)


##


@ta.overload
def ref(obj: T) -> Ref[T, K]:
    ...


@ta.overload
def ref(cls: type[T], k: Key[K]) -> Ref[T, K]:
    ...


@ta.overload
def ref(cls: type[T], k: K) -> Ref[T, K]:
    ...


def ref(obj, *args):
    if not args:
        check.not_isinstance(obj, type)
        return _DirectRef(obj)
    else:
        check.isinstance(obj, type)
        [k] = args
        return _LazyRef(obj, key(k))


##


_REF_TYPES: tuple[type[Ref], ...] = (
    _DirectRef,
    _LazyRef,
)



# ruff: noqa: UP006 UP007 UP045
# @omlish-lite
import concurrent.futures as cf
import dataclasses as dc
import socket
import typing as ta

from ...logs.protocols import LoggerLike
from ..addresses import SocketAndAddress
from ..io import close_socket_immediately
from .types import SocketHandler
from .types import SocketHandler_


##


@dc.dataclass(frozen=True)
class StandardSocketHandler(SocketHandler_):
    handler: SocketHandler

    timeout: ta.Optional[float] = None

    # http://bugs.python.org/issue6192
    # TODO: https://eklitzke.org/the-caveats-of-tcp-nodelay
    enable_nodelay: bool = False

    no_close: bool = False

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            if self.timeout is not None:
                conn.socket.settimeout(self.timeout)

            if self.enable_nodelay:
                conn.socket.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, True)

            self.handler(conn)

        finally:
            close_socket_immediately(conn.socket)


##


@dc.dataclass(frozen=True)
class CallbackWrappedSocketHandler(SocketHandler_):
    handler: SocketHandler

    before_handle: ta.Optional[SocketHandler] = None
    after_handle: ta.Optional[SocketHandler] = None

    # Return True if suppress like __exit__
    on_error: ta.Optional[ta.Callable[[SocketAndAddress, Exception], bool]] = None

    finally_: ta.Optional[SocketHandler] = None

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            if (before_handle := self.before_handle) is not None:
                before_handle(conn)

            self.handler(conn)

        except Exception as e:
            if (on_error := self.on_error) is not None and on_error(conn, e):
                pass
            else:
                raise

        else:
            if (after_handle := self.after_handle) is not None:
                after_handle(conn)

        finally:
            if (finally_ := self.finally_) is not None:
                finally_(conn)


##


@dc.dataclass(frozen=True)
class SocketWrappingSocketHandler(SocketHandler_):
    handler: SocketHandler
    wrapper: ta.Callable[[SocketAndAddress], SocketAndAddress]

    def __call__(self, conn: SocketAndAddress) -> None:
        wrapped_conn = self.wrapper(conn)
        self.handler(wrapped_conn)


##


@dc.dataclass(frozen=True)
class ExecutorSocketHandler(SocketHandler_):
    handler: SocketHandler
    executor: cf.Executor

    def __call__(self, conn: SocketAndAddress) -> None:
        self.executor.submit(self.handler, conn)


##


@dc.dataclass(frozen=True)
class ExceptionLoggingSocketHandler(SocketHandler_):
    handler: SocketHandler
    log: LoggerLike

    ignored: ta.Optional[ta.Container[ta.Type[Exception]]] = None

    def __call__(self, conn: SocketAndAddress) -> None:
        try:
            return self.handler(conn)

        except Exception as e:  # noqa
            if (ignored := self.ignored) is None or type(e) not in ignored:
                self.log.exception('Error in handler %r for conn %r', self.handler, conn)

            raise



import abc
import typing as ta

from .. import dataclasses as dc
from .. import lang
from .mappers import Mapper
from .snaps import Snap


##


class Store(lang.Abstract):
    @abc.abstractmethod
    def fetch(self, m: Mapper, k: ta.Any) -> Snap | None:
        raise NotImplementedError

    #

    @abc.abstractmethod
    def lookup(self, m: Mapper, where: ta.Mapping[str, ta.Any]) -> ta.Sequence[Snap]:
        raise NotImplementedError

    #

    @ta.final
    @dc.dataclass(frozen=True, kw_only=True)
    class FlushBatch:
        insert: ta.Sequence[Snap] | None = None
        update: ta.Sequence[tuple[ta.Any, Snap]] | None = None
        delete: ta.Sequence[ta.Any] | None = None  # noqa

    @ta.final
    @dc.dataclass(frozen=True, kw_only=True)
    class FlushResult:
        inserted_auto_keys: ta.Mapping[ta.Any, ta.Any] | None = None

    @abc.abstractmethod
    def flush(self, m: Mapper, b: FlushBatch) -> FlushResult:
        raise NotImplementedError



# ruff: noqa: SLF001
# MIT License
#
# Copyright (c) 2022 Gabriel Ochsenhofer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import random
import typing as ta

from ... import lang


T = ta.TypeVar('T')

Comparer: ta.TypeAlias = ta.Callable[[T, T], int]


##


@ta.final
class TreapNode(ta.Generic[T]):
    __slots__ = ('_value', '_priority', '_left', '_right', '_count')

    def __init__(
            self,
            *,
            _value: T,
            _priority: int,
            _left: ta.Optional['TreapNode[T]'],
            _right: ta.Optional['TreapNode[T]'],
    ) -> None:
        self._value, self._priority, self._left, self._right = _value, _priority, _left, _right
        self._set_count()

    _count: int

    def _set_count(self) -> None:
        self._count = (
            1 +
            (self._left._count if self._left is not None else 0) +
            (self._right._count if self._right is not None else 0)
        )

    def __repr__(self) -> str:
        return f'TreapNode(value={self._value!r}, priority={self._priority!r}, count={self._count!r})'

    @property
    def value(self) -> T:
        return self._value

    @property
    def priority(self) -> int:
        return self._priority

    @property
    def left(self) -> ta.Optional['TreapNode[T]']:
        return self._left

    @property
    def right(self) -> ta.Optional['TreapNode[T]']:
        return self._right

    @property
    def count(self) -> int:
        return self._count

    def __iter__(self) -> ta.Iterator[T]:
        if self._left is not None:
            yield from self._left
        yield self._value
        if self._right is not None:
            yield from self._right


def _new_priority() -> int:
    return random.getrandbits(32)


def new(
        value: T,
        *,
        priority: int | None = None,
) -> TreapNode[T]:
    return TreapNode(_value=value, _priority=priority if priority is not None else _new_priority(), _left=None, _right=None)  # noqa


def find(
        n: TreapNode[T] | None,
        v: T,
        c: Comparer[T] | None,
) -> TreapNode[T] | None:
    if c is None:
        c = lang.cmp

    while True:
        if n is None:
            return None

        diff = c(n._value, v)  # noqa

        if diff == 0:
            return n

        elif diff < 0:
            n = n._right  # noqa

        else:
            n = n._left  # noqa


def place(
        n: TreapNode[T] | None,
        v: T,
        c: Comparer[T] | None,
        desc: bool,
) -> list[TreapNode[T]]:
    if c is None:
        c = lang.cmp

    ret: list[TreapNode[T]] = []

    while True:
        if n is None:
            break

        diff = c(n._value, v)  # noqa

        if diff == 0:
            ret.append(n)
            break

        elif diff < 0:
            if desc:
                ret.append(n)
            n = n._right  # noqa

        else:
            if not desc:
                ret.append(n)
            n = n._left  # noqa

    return ret


def union(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
        c: Comparer[T] | None,
        overwrite: bool,
) -> TreapNode[T] | None:
    if n is None:
        return other

    if other is None:
        return n

    if n._priority < other._priority:
        # If we are unioning Tree A (target) and Tree B (source), and overwrite=True (meaning B's values should replace
        # A's on a collision):
        # - If A._priority < B._priority (meaning A becomes the parent of B), the code naturally extracts B's value as
        #   dupe, and value = dupe._value overwrites the root A's value. This is correct.
        # - If B._priority < A._priority (meaning B becomes the parent of A), you swap them so n is now B.
        # - Because n is now B, its baseline value is B's value.
        # - If you didn't flip overwrite, the code would see overwrite=True, find A's value as the dupe, and overwrite
        #   B's value with A's value - which is exactly backward!
        # - By flipping overwrite = not overwrite (making it False), you prevent A from overwriting B, thus preserving
        #   B's value as the root.
        other, n, overwrite = n, other, not overwrite

    left, dupe, right = split(other, n._value, c)

    value = n._value
    if overwrite and dupe is not None:
        value = dupe._value

    left = union(n._left, left, c, overwrite)
    right = union(n._right, right, c, overwrite)

    return TreapNode(_value=value, _priority=n._priority, _left=left, _right=right)


def split(
        n: TreapNode[T] | None,
        v: T,
        c: Comparer[T] | None,
) -> tuple[
        TreapNode[T] | None,
        TreapNode[T] | None,
        TreapNode[T] | None,
]:
    if n is None:
        return None, None, None

    if c is None:
        c = lang.cmp

    diff = c(n._value, v)

    if diff < 0:
        # n._value < v -> n belongs in the left split. We must recursively split n's right child.
        left_split, dupe, right_split = split(n._right, v, c)

        # Rebuild n with its original left child, and the new left_split as its right child. _set_count() is called
        # safely here because left_split is fully formed.
        new_left = TreapNode(
            _value=n._value,
            _priority=n._priority,
            _left=n._left,
            _right=left_split,
        )
        return new_left, dupe, right_split

    elif diff > 0:
        # n._value > v -> n belongs in the right split. We must recursively split n's left child.
        left_split, dupe, right_split = split(n._left, v, c)

        # Rebuild n with the new right_split as its left child, and its original right child.
        new_right = TreapNode(
            _value=n._value,
            _priority=n._priority,
            _left=right_split,
            _right=n._right,
        )
        return left_split, dupe, new_right

    else:
        # diff == 0 -> We found the split point exactly. Isolate the duplicate node (no children).
        dupe = TreapNode(_value=n._value, _priority=n._priority, _left=None, _right=None)
        return n._left, dupe, n._right


def intersect(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
        c: Comparer[T] | None,
) -> TreapNode[T] | None:
    if n is None or other is None:
        return None

    if n._priority < other._priority:
        n, other = other, n

    left, found, right = split(other, n._value, c)

    left = intersect(n._left, left, c)
    right = intersect(n._right, right, c)

    if found is None:
        # TODO: use a destructive join as both left/right are copies
        return _join(left, right)

    return TreapNode(_value=n._value, _priority=n._priority, _left=left, _right=right)


def delete(n: TreapNode[T] | None, v: T, c: Comparer[T]) -> TreapNode[T] | None:
    left, _, right = split(n, v, c)
    return _join(left, right)


def diff(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
        c: Comparer[T] | None,
) -> TreapNode[T] | None:
    if n is None or other is None:
        return n

    if n._priority >= other._priority:
        left, dupe, right = split(other, n._value, c)

        left, right = diff(n._left, left, c), diff(n._right, right, c)

        if dupe is not None:
            return _join(left, right)

        return TreapNode(_value=n._value, _priority=n._priority, _left=left, _right=right)

    left, _, right = split(n, other._value, c)

    left = diff(left, other._left, c)
    right = diff(right, other._right, c)

    return _join(left, right)


def _join(
        n: TreapNode[T] | None,
        other: TreapNode[T] | None,
) -> TreapNode[T] | None:
    if n is None:
        return other
    if other is None:
        return n

    if n._priority >= other._priority:
        # 'n' is the root. Since all keys in 'n' < all keys in 'other', 'other' must be merged into 'n's right side.
        right_joined = _join(n._right, other)

        return TreapNode(
            _value=n._value,
            _priority=n._priority,
            _left=n._left,
            _right=right_joined,
        )

    else:
        # 'other' is the root. Since all keys in 'n' < all keys in 'other', 'n' must be merged into 'other's left side.
        left_joined = _join(n, other._left)

        return TreapNode(
            _value=other._value,
            _priority=other._priority,
            _left=left_joined,
            _right=other._right,
        )



"""
TODO:
 - @lang.copy_type
"""
import dataclasses as dc
import typing as ta

from ... import cached
from ... import check
from ... import lang
from ...lite.dataclasses import install_dataclass_filtered_repr
from ..api.naming import Naming
from ..api.types import Marshaler
from ..api.types import MarshalerFactory
from ..api.types import Unmarshaler
from ..api.types import UnmarshalerFactory


T = ta.TypeVar('T')


##


@ta.final
@install_dataclass_filtered_repr('omit_falsey')
@dc.dataclass(frozen=True, kw_only=True)
class FieldOptions(lang.Final):
    """
    Unified field options - all configuration for a single field's marshaling/unmarshaling.

    This is the single source of truth for field configuration, merging what used to be split between FieldOptions and
    FieldMetadata in the old design.
    """

    ##
    # Naming

    name: str | None = None
    alts: ta.Iterable[str] | None = None

    ##
    # Behavior options

    omit_if: ta.Callable[[ta.Any], bool] | None = None

    default: lang.Maybe[ta.Any] | None = None

    embed: bool | None = None

    generic_replace: bool | None = None

    no_marshal: bool | None = None
    no_unmarshal: bool | None = None

    ##
    # Custom handlers

    marshaler: Marshaler | None = None
    marshaler_factory: MarshalerFactory | None = None
    marshal_as: ta.Any | None = None

    unmarshaler: Unmarshaler | None = None
    unmarshaler_factory: UnmarshalerFactory | None = None
    unmarshal_as: ta.Any | None = None

    ##
    # Validation

    def __post_init__(self) -> None:
        check.isinstance(self.default, (lang.Maybe, None))

        check.isinstance(self.marshaler, (Marshaler, None))
        check.isinstance(self.unmarshaler, (Unmarshaler, None))

        for mutex_atts in [
            ('marshaler', 'marshaler_factory', 'marshal_as'),
            ('unmarshaler', 'unmarshaler_factory', 'unmarshal_as'),
        ]:
            set_atts = [a for a in mutex_atts if getattr(self, a) is not None]
            if len(set_atts) > 1:
                raise ValueError(f'Cannot set {set_atts} at the same time')

    ##
    # Merging

    def merge(self, *overrides: ta.Optional['FieldOptions']) -> 'FieldOptions':
        kw: dict[str, ta.Any] = {}
        for obj in [self, *overrides]:
            if obj is None:
                continue

            for fld in _field_options_fields():
                if (fv := getattr(obj, fld.name)) is None:
                    continue

                kw[fld.name] = fv

        if not kw:
            return self
        return FieldOptions(**kw)


@lang.cached_function
def _field_options_fields() -> ta.Sequence[dc.Field]:
    return dc.fields(FieldOptions)


DEFAULT_FIELD_OPTIONS = FieldOptions()


##


@ta.final
@dc.dataclass(frozen=True, kw_only=True)
class ObjectSpecials(lang.Final):
    """Special field names for an object."""

    unknown: str | None = None
    source: str | None = None

    @cached.property
    def set(self) -> frozenset[str]:
        return frozenset(v for v in dc.asdict(self).values() if v is not None)


@ta.final
@install_dataclass_filtered_repr('omit_falsey')
@dc.dataclass(frozen=True, kw_only=True)
class ObjectOptions(lang.Final):
    """Object-level marshaling options."""

    ##
    # Naming

    field_naming: Naming | None = None

    ##
    # Behavior options

    ignore_unknown: bool | None = None

    unwrap_if_single_field: ta.Literal['marshal', 'unmarshal', True, None] = None

    ##
    # Special fields

    unknown_field: str | None = None
    source_field: str | None = None

    @cached.property
    def specials(self) -> ObjectSpecials:
        return ObjectSpecials(
            unknown=self.unknown_field,
            source=self.source_field,
        )

    ##
    # Field defaults

    field_defaults: FieldOptions = DEFAULT_FIELD_OPTIONS

    ##
    # Merging

    def merge(self, *overrides: ta.Optional['ObjectOptions']) -> 'ObjectOptions':
        kw: dict[str, ta.Any] = {}
        for obj in [self, *overrides]:
            if obj is None:
                continue

            for fld in _object_options_fields():
                if (fv := getattr(obj, fld.name)) is None:
                    continue

                if fld.type is FieldOptions:
                    fv = kw.get(fld.name, DEFAULT_FIELD_OPTIONS).merge(fv)

                kw[fld.name] = fv

        if not kw:
            return self
        return ObjectOptions(**kw)


@lang.cached_function
def _object_options_fields() -> ta.Sequence[dc.Field]:
    return dc.fields(ObjectOptions)


DEFAULT_OBJECT_OPTIONS = ObjectOptions()



"""
Note that this is similar to but different from `omlish.c3` - this is for validation and debugging of real. concrete,
intended-to-be-final class bases, whereas `omlish.c3` is for computing mro of arbitrary otherwise unrelated types,
including virtual ABC's (primarily for dispatch-like code).
"""
import typing as ta


##


def _resolve_bases(bases: tuple[ta.Any, ...]) -> tuple[tuple[type, ...], tuple[int, ...]]:
    out_bases: list[type] = []
    out_orig_idxs: list[int] = []

    for i, base in enumerate(bases):
        entries = _resolve_one_base_entries(base, bases)

        for e in entries:
            if not isinstance(e, type):
                raise TypeError(f'resolved base entry at bases[{i}] is not a type: {e!r}')
            out_bases.append(e)
            out_orig_idxs.append(i)

    return tuple(out_bases), tuple(out_orig_idxs)


def _resolve_one_base_entries(base: ta.Any, bases: tuple[ta.Any, ...]) -> tuple[ta.Any, ...]:
    if isinstance(base, type):
        return (base,)

    mro_entries = getattr(base, '__mro_entries__', None)
    if mro_entries is not None:
        entries = mro_entries(bases)
        if not isinstance(entries, tuple):
            raise TypeError(f'__mro_entries__ for {base!r} returned non-tuple {entries!r}')
        return entries

    raise TypeError(f'base {base!r} is neither a type nor provides __mro_entries__')


class _Seq:
    def __init__(
        self,
        items: list[type],
        *,
        origin_idxs: list[int] | None = None,
        fixed_origin_idx: int | None = None,
    ) -> None:
        self.items = items
        self._origin_idxs = origin_idxs
        self._fixed_origin_idx = fixed_origin_idx

    def head(self) -> type:
        return self.items[0]

    def head_origin_idx(self) -> int:
        if self._origin_idxs is not None:
            return self._origin_idxs[0]
        if self._fixed_origin_idx is None:
            raise RuntimeError(f'head_origin_idx called on {self!r} with no origin_idxs or fixed_origin_idx')
        return self._fixed_origin_idx

    def iter_pairs(self, orig_bases: tuple[ta.Any, ...]) -> tuple[tuple[ta.Any, type], ...]:
        if self._origin_idxs is not None:
            return tuple((orig_bases[oi], ty) for oi, ty in zip(self._origin_idxs, self.items))
        else:
            if self._fixed_origin_idx is None:
                raise RuntimeError(f'iter_pairs called on {self!r} with no origin_idxs or fixed_origin_idx')
            ob = orig_bases[self._fixed_origin_idx]
            return tuple((ob, ty) for ty in self.items)

    def pop_head_if(self, candidate: type) -> None:
        if self.items and self.items[0] is candidate:
            self.items.pop(0)
            if self._origin_idxs is not None:
                self._origin_idxs.pop(0)

    def tail_contains(self, ty: type) -> bool:
        return ty in self.items[1:]


def _merge_mro_or_raise(
    *,
    orig_bases: tuple[ta.Any, ...],
    expanded_bases: tuple[type, ...],
    expanded_orig_idxs: tuple[int, ...],
) -> None:
    seqs: list[_Seq] = []

    for base, orig_idx in zip(expanded_bases, expanded_orig_idxs):
        seqs.append(_Seq(list(base.__mro__), fixed_origin_idx=orig_idx))

    seqs.append(
        _Seq(list(expanded_bases), origin_idxs=list(expanded_orig_idxs)),
    )

    while True:
        seqs = [s for s in seqs if s.items]
        if not seqs:
            return

        candidate: type | None = None
        for s in seqs:
            h = s.head()
            if all(not other.tail_contains(h) for other in seqs):
                candidate = h
                break

        if candidate is None:
            conflicts = _build_conflicts(orig_bases=orig_bases, seqs=seqs)
            remaining_sequences = tuple(s.iter_pairs(orig_bases) for s in seqs)
            raise MroError(
                bases=orig_bases,
                expanded_bases=expanded_bases,
                conflicts=conflicts,
                remaining_sequences=remaining_sequences,
            )

        for s in seqs:
            s.pop_head_if(candidate)


def _build_conflicts(
    *,
    orig_bases: tuple[ta.Any, ...],
    seqs: list[_Seq],
) -> dict[str, tuple[ta.Any, ...]]:
    out: dict[str, tuple[ta.Any, ...]] = {}
    used_keys: set[str] = set()

    for s in seqs:
        h = s.head()
        blockers: list[int] = []

        for other in seqs:
            if other is s:
                continue
            if other.tail_contains(h):
                blocker_idx = other.head_origin_idx()
                if blocker_idx not in blockers:
                    blockers.append(blocker_idx)

        if not blockers:
            continue

        involved_idxs = [s.head_origin_idx(), *blockers]
        key = _dedupe_key(_type_name(h), used_keys)
        out[key] = tuple(orig_bases[i] for i in involved_idxs)

    return out


def _type_name(ty: type) -> str:
    mod = getattr(ty, '__module__', None)
    qn = getattr(ty, '__qualname__', None)
    if mod and qn:
        return f'{mod}.{qn}'
    n = getattr(ty, '__name__', None)
    if n:
        return n
    return repr(ty)


def _dedupe_key(key: str, used: set[str]) -> str:
    if key not in used:
        used.add(key)
        return key

    i = 2
    while True:
        k = f'{key}#{i}'
        if k not in used:
            used.add(k)
            return k
        i += 1


#


class MroError(TypeError):
    """
    Raised when the given base list cannot be merged into a consistent C3 MRO.

    Attributes:
        bases:
            The exact original objects passed to try_mro(...).

        expanded_bases:
            The flattened type bases after applying __mro_entries__.

        conflicts:
            Mapping from a problematic remaining merge-head name to the ordered original base-list objects involved in
            that conflict.

            The first item is the original base object whose sequence currently proposes that head; subsequent items are
            the original base objects whose current heads force that head to come later.

            These are the *exact* original base-list objects, not their expanded types, and may therefore be arbitrary /
            unhashable objects.

        remaining_sequences:
            The C3 merge sequences at the exact failure point, rendered in terms of the original base-list objects. Each
            entry is a tuple of pairs:
                ((original_base_object, remaining_type), ...)
            where original_base_object is the exact object from the input base list responsible for that remaining type
            position.
    """

    def __init__(
            self,
            *,
            bases: tuple[ta.Any, ...],
            expanded_bases: tuple[type, ...],
            conflicts: dict[str, tuple[ta.Any, ...]],
            remaining_sequences: tuple[tuple[tuple[ta.Any, type], ...], ...],
    ) -> None:
        self.bases = bases
        self.expanded_bases = expanded_bases
        self.conflicts = conflicts
        self.remaining_sequences = remaining_sequences
        super().__init__(self._format_message())

    def _format_message(self) -> str:
        lines = [
            'Cannot create a consistent method resolution order (MRO) for bases.',
            '',
            'Original bases:',
        ]
        for i, b in enumerate(self.bases):
            lines.append(f'  [{i}] {b!r}')

        lines.append('')
        lines.append('Expanded bases:')
        for i, b in enumerate(self.expanded_bases):
            lines.append(f'  [{i}] {_type_name(b)}')

        lines.append('')
        lines.append('Conflicts:')
        for k, vs in self.conflicts.items():
            rendered = ', '.join(repr(v) for v in vs)
            lines.append(f'  {k}: ({rendered})')

        lines.append('')
        lines.append('Remaining sequences at failure:')
        for i, seq in enumerate(self.remaining_sequences):
            rendered = ', '.join(f'({ob!r}, {_type_name(ty)})' for ob, ty in seq)
            lines.append(f'  [{i}] [{rendered}]')

        return '\n'.join(lines)


def try_mro(*bases: ta.Any) -> tuple:
    """
    Validate whether `class X(*bases): ...` would have a consistent MRO.

    This performs C3 linearization on the resolved bases without actually creating a class. It preserves a back-map to
    the *original* base-list objects, including generic aliases and arbitrary __mro_entries__ expanders.

    Returns:
        The given bases if the MRO is consistent.

    Raises:
        MroError:
            If the bases cannot be merged into a consistent MRO.
        TypeError:
            If a base is invalid in a way unrelated to MRO consistency (for example, not a type and with no
            __mro_entries__).
    """

    orig_bases = tuple(bases)
    expanded_bases, expanded_orig_idxs = _resolve_bases(orig_bases)

    _merge_mro_or_raise(
        orig_bases=orig_bases,
        expanded_bases=expanded_bases,
        expanded_orig_idxs=expanded_orig_idxs,
    )

    return bases



import typing as ta
import weakref


T = ta.TypeVar('T')
V = ta.TypeVar('V')


##


class TypeMap(ta.Generic[T]):
    def __init__(self, items: ta.Iterable[T] = ()) -> None:
        super().__init__()

        self._items = list(items)
        dct: dict[type, ta.Any] = {}
        for item in items:
            if (ty := type(item)) in dct:
                raise ValueError(ty)
            dct[ty] = item
        self._dct = dct

    @classmethod
    def of(cls, items: ta.Iterable[T]) -> 'TypeMap[T]':
        if isinstance(items, TypeMap):
            return items
        return cls(items)

    @property
    def items(self) -> ta.Sequence[T]:
        return self._items

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> ta.Iterator[T]:
        return iter(self._items)

    def __contains__(self, ty: type[T]) -> bool:
        return ty in self._items

    def get(self, ty: type[T]) -> T | None:
        return self._dct.get(ty)

    def __getitem__(self, ty: type[T]) -> T:
        return self._dct[ty]

    _any_dct: dict[type | tuple[type, ...], tuple[T, ...]]

    def get_any(self, cls: type | tuple[type, ...]) -> ta.Sequence[T]:
        try:
            any_dct = self._any_dct
        except AttributeError:
            any_dct = {}
            self._any_dct = any_dct

        try:
            return any_dct[cls]
        except KeyError:
            pass

        ret = tuple(tv for tv in self if isinstance(tv, cls))
        any_dct[cls] = ret
        return ret


class DynamicTypeMap(ta.Generic[V]):
    def __init__(self, items: ta.Iterable[V] = (), *, weak: bool = False) -> None:
        super().__init__()

        self._items = list(items)
        self._weak = bool(weak)

        self._cache: ta.MutableMapping[type, ta.Any] = weakref.WeakKeyDictionary() if weak else {}

    @property
    def items(self) -> ta.Sequence[V]:
        return self._items

    @property
    def is_weak(self) -> bool:
        return self._weak

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> ta.Iterator[V]:
        return iter(self._items)

    def __getitem__(self, ty: type[T]) -> ta.Sequence[T]:
        try:
            return self._cache[ty]
        except KeyError:
            ret = []
            for item in self._items:
                if isinstance(item, ty):
                    ret.append(item)
            self._cache[ty] = ret
            return ret



"""
This package (together with related `api` modules in sibling packages like `objects`) is the lightweight api intended
to be imported for marshal customization at user definitions. As such it must avoid heavy imports, even if those imports
would otherwise result in faster execution. Notably:

- Use stdlib `dataclasses` not `omlish.dataclasses`
- Use dumb objects for config, not `omlish.typedvalues`
"""



import operator
import typing as ta

from ... import check
from ..mappings import IterItemsViewMapping
from ..mappings import IterValuesViewMapping
from ..persistent import PersistentMapping


try:
    from . import _hamt  # type: ignore
except ImportError:
    pass


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


@ta.final
class HamtMap(
    IterValuesViewMapping[K, V],
    IterItemsViewMapping[K, V],
    PersistentMapping[K, V],
):
    def __init__(self, *, _h: ta.Any | None = None) -> None:
        self._h = _h if _h is not None else _hamt.new()

    def __len__(self) -> int:
        return _hamt.len(self._h)

    def __contains__(self, item: K) -> bool:  # type: ignore[override]
        return _hamt.find(self._h, item) is not None

    def __getitem__(self, item: K) -> V:
        if (v := _hamt.find(self._h, item)) is None:
            raise KeyError(item)
        return v

    def __iter__(self) -> ta.Iterator[K]:
        return _hamt.iter_keys(self._h)

    def iteritems(self) -> ta.Iterator[tuple[K, V]]:
        return _hamt.iter_items(self._h)

    def itervalues(self) -> ta.Iterator[V]:
        return map(operator.itemgetter(1), self.iteritems())

    def with_(self, k: K, v: V) -> ta.Self:
        return HamtMap(_h=_hamt.assoc(self._h, k, check.not_none(v)))

    def without(self, k: K) -> ta.Self:
        if (h2 := _hamt.without(h1 := self._h, k)) is h1:
            return self
        return HamtMap(_h=h2)

    def default(self, k: K, v: V) -> ta.Self:
        if _hamt.find(self._h, k) is not None:
            return self
        return self.with_(k, v)


def new_hamt_map(
        items: ta.Iterable[tuple[K, V]] | None = None,
) -> HamtMap[K, V]:
    m: HamtMap[K, V] = HamtMap()
    if items is not None:
        for k, v in items:
            m = m.with_(k, v)
    return m


def is_hamt_available() -> bool:
    try:
        _hamt  # noqa
    except NameError:
        return False
    else:
        return True



"""
TODO:
 - heterogeneous tuples
"""
import collections.abc
import functools
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.options import Options
from ..api.types import Marshaler
from ..api.types import Unmarshaler
from ..api.values import Value
from ..factories.method import MarshalerFactoryMethodClass
from ..factories.method import UnmarshalerFactoryMethodClass
from .api import DefaultIterableConstructors


##


DEFAULT_ITERABLE_CONCRETE_TYPES: dict[type[collections.abc.Iterable], type[collections.abc.Iterable]] = {
    collections.abc.Iterable: tuple,  # type: ignore
    collections.abc.Sequence: tuple,  # type: ignore
    collections.abc.MutableSequence: list,  # type: ignore
}


def get_default_iterable_constructor(
        cls: type,
        options: Options | None = None,
) -> ta.Callable[[collections.abc.Iterable], ta.Any]:
    if options is not None and (opt := options.get(DefaultIterableConstructors)) is not None:
        opt = check.isinstance(opt, DefaultIterableConstructors)
        o_ctor: ta.Any = None
        if cls == collections.abc.Iterable:
            o_ctor = opt.iterable
        elif cls == collections.abc.Sequence:
            o_ctor = opt.sequence
        elif cls == collections.abc.MutableSequence:
            o_ctor = opt.mutable_sequence
        if o_ctor is not None:
            return o_ctor

    return DEFAULT_ITERABLE_CONCRETE_TYPES.get(cls, cls)  # noqa


#


@dc.dataclass(frozen=True)
class IterableMarshaler(Marshaler):
    e: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Iterable) -> Value:
        return list(map(functools.partial(self.e.marshal, ctx), o))


class IterableMarshalerFactory(MarshalerFactoryMethodClass):
    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_generic(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable)):
            return None
        return lambda: IterableMarshaler(ctx.make_marshaler(check.single(rty.args)))

    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_concrete(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, collections.abc.Iterable)):
            return None
        return lambda: IterableMarshaler(ctx.make_marshaler(ta.Any))


#


@dc.dataclass(frozen=True)
class IterableUnmarshaler(Unmarshaler):
    cls: type
    e: Unmarshaler

    ctor: ta.Callable[[collections.abc.Iterable], ta.Any] | None = None

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Iterable:
        if (ctor := self.ctor) is None:
            ctor = get_default_iterable_constructor(self.cls, ctx.options)
        return ctor(map(functools.partial(self.e.unmarshal, ctx), check.isinstance(v, collections.abc.Iterable)))


class IterableUnmarshalerFactory(UnmarshalerFactoryMethodClass):
    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_generic(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Iterable)):
            return None
        return lambda: IterableUnmarshaler(rty.cls, ctx.make_unmarshaler(check.single(rty.args)))  # noqa

    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_concrete(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, collections.abc.Iterable)):
            return None
        return lambda: IterableUnmarshaler(check.isinstance(rty, type), ctx.make_unmarshaler(ta.Any))



import typing as ta


_comparison: ta.Any
try:
    from . import _comparison  # type: ignore
except ImportError:
    _comparison = None


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


def cmp(l: ta.Any, r: ta.Any) -> int:
    return int(l > r) - int(l < r)


def key_cmp(fn: ta.Callable[[K, K], int] | None = None) -> ta.Callable[[tuple[K, V], tuple[K, V]], int]:
    if fn is None:
        fn = cmp
    return lambda t0, t1: fn(t0[0], t1[0])


if _comparison is not None:
    globals().update({a: getattr(_comparison, a) for a in [
        'cmp',
        'key_cmp',
    ]})


##


class InfinityType:
    def __repr__(self) -> str:
        return 'Infinity'

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: ta.Any) -> bool:
        return False

    def __le__(self, other: ta.Any) -> bool:
        return False

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __gt__(self, other: ta.Any) -> bool:
        return True

    def __ge__(self, other: ta.Any) -> bool:
        return True

    def __neg__(self: ta.Any) -> 'NegativeInfinityType':
        return NegativeInfinity


Infinity = InfinityType()


class NegativeInfinityType:
    def __repr__(self) -> str:
        return '-Infinity'

    def __hash__(self) -> int:
        return hash(repr(self))

    def __lt__(self, other: ta.Any) -> bool:
        return True

    def __le__(self, other: ta.Any) -> bool:
        return True

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __gt__(self, other: ta.Any) -> bool:
        return False

    def __ge__(self, other: ta.Any) -> bool:
        return False

    def __neg__(self: ta.Any) -> InfinityType:
        return Infinity


NegativeInfinity = NegativeInfinityType()



# MIT License
#
# Copyright (c) 2022 Gabriel Ochsenhofer
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import abc
import typing as ta

from ... import lang
from ..intersections import PersistentSortedMapping
from ..mappings import IterItemsViewMapping
from ..mappings import IterValuesViewMapping
from ..mappings import iteritems_itervalues
from ..mappings import map_contains
from . import treap


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


@ta.final
class TreapMap(
    IterValuesViewMapping[K, V],
    IterItemsViewMapping[K, V],
    PersistentSortedMapping[K, V],
    ta.Mapping[K, V],
    ta.Generic[K, V],
):
    __slots__ = ('_n', '_c')

    def __init__(
            self,
            *,
            _n: treap.TreapNode[tuple[K, V]] | None,
            _c: treap.Comparer[tuple[K, V]],
    ) -> None:
        self._n, self._c = _n, _c

    def __len__(self) -> int:
        return self._n.count if self._n is not None else 0

    def __getitem__(self, item: K) -> V:
        n = treap.find(self._n, (item, None), self._c)  # type: ignore
        if n is None:
            raise KeyError(item)
        return n.value[1]  # type: ignore[return-value]

    def __iter__(self) -> ta.Iterator[K]:
        i = self.iteritems()
        while i.has_next():
            yield i.next()[0]

    __contains__ = map_contains

    itervalues = iteritems_itervalues

    def iteritems(self) -> 'TreapMapIterator[K, V]':
        i = TreapMapIterator(_st=[], _n=self._n)
        while (n := i._n) is not None and n.left is not None:  # noqa
            i._st.append(n)  # noqa
            i._n = n.left  # noqa
        return i

    def items_desc(self) -> 'TreapMapReverseIterator[K, V]':
        i = TreapMapReverseIterator(_st=[], _n=self._n)
        while (n := i._n) is not None and n.right is not None:  # noqa
            i._st.append(n)  # noqa
            i._n = n.right  # noqa
        return i

    def items_from(self, k: K) -> 'TreapMapIterator[K, V]':
        lst: list = treap.place(self._n, (k, None), self._c, False)  # type: ignore
        i = TreapMapIterator(_st=lst, _n=lst.pop() if lst else None)
        return i

    def items_from_desc(self, k: K) -> 'TreapMapReverseIterator[K, V]':
        lst: list = treap.place(self._n, (k, None), self._c, True)  # type: ignore
        i = TreapMapReverseIterator(_st=lst, _n=lst.pop() if lst else None)
        return i

    def with_(self, k: K, v: V) -> 'TreapMap[K, V]':
        node = treap.new((k, v))
        n = treap.union(self._n, node, self._c, True)  # type: ignore[arg-type]
        return TreapMap(_n=n, _c=self._c)

    def without(self, k: K) -> 'TreapMap[K, V]':
        n = treap.delete(self._n, (k, None), self._c)  # type: ignore
        return TreapMap(_n=n, _c=self._c)  # type: ignore[arg-type]

    def default(self, k: K, v: V) -> 'TreapMap[K, V]':
        try:
            self[k]  # noqa
        except KeyError:
            return self.with_(k, v)
        else:
            return self


def new_treap_map(
        items: ta.Iterable[tuple[K, V]] | None = None,
        *,
        cmp: ta.Callable[[K, K], int] | None = None,
) -> TreapMap[K, V]:
    m: TreapMap[K, V] = TreapMap(_n=None, _c=lang.key_cmp(cmp))
    if items is not None:
        for k, v in items:
            m = m.with_(k, v)
    return m


##


class BaseTreapMapIterator(lang.Abstract, ta.Iterator[tuple[K, V]], ta.Generic[K, V]):
    __slots__ = ('_st', '_n')

    def __init__(
            self,
            *,
            _st: list[treap.TreapNode[tuple[K, V]]],
            _n: treap.TreapNode[tuple[K, V]] | None,
    ) -> None:
        self._st, self._n = _st, _n

    def __iter__(self) -> ta.Self:
        return self

    def __next__(self) -> tuple[K, V]:
        return self.next()

    def has_next(self) -> bool:
        return self._n is not None

    @abc.abstractmethod
    def next(self) -> tuple[K, V]:
        raise NotImplementedError


@ta.final
class TreapMapIterator(BaseTreapMapIterator[K, V]):
    __slots__ = BaseTreapMapIterator.__slots__  # noqa

    def next(self) -> tuple[K, V]:
        n = self._n
        if n is None:
            raise StopIteration
        if n.right is not None:
            self._n = n.right
            while self._n.left is not None:
                self._st.append(self._n)
                self._n = self._n.left
        elif len(self._st) > 0:
            self._n = self._st[-1]
            self._st.pop()
        else:
            self._n = None
        return n.value


@ta.final
class TreapMapReverseIterator(BaseTreapMapIterator[K, V]):
    __slots__ = BaseTreapMapIterator.__slots__  # noqa

    def next(self) -> tuple[K, V]:
        n = self._n
        if n is None:
            raise StopIteration
        if n.left is not None:
            self._n = n.left
            while self._n.right is not None:
                self._st.append(self._n)
                self._n = self._n.right
        elif len(self._st) > 0:
            self._n = self._st[-1]
            self._st.pop()
        else:
            self._n = None
        return n.value



import abc
import types
import typing as ta

from ..lite.abstract import Abstract


T = ta.TypeVar('T')


##


class Cell(ta.Protocol[T]):
    def __call__(self) -> T: ...

    def set(self, v: T) -> None: ...

    #

    def __bool__(self) -> ta.NoReturn: ...

    def __hash__(self) -> ta.NoReturn: ...

    def __eq__(self, other: object) -> ta.NoReturn: ...

    def __ne__(self, other: object) -> ta.NoReturn: ...


class BaseCell(Abstract, ta.Generic[T]):
    @abc.abstractmethod
    def __call__(self) -> T:
        raise NotImplementedError

    @abc.abstractmethod
    def set(self, v: T) -> None:
        raise NotImplementedError

    #

    @ta.final
    def __bool__(self):
        raise TypeError(self)

    @ta.final
    def __hash__(self):
        raise TypeError(self)

    @ta.final
    def __eq__(self, other):
        raise TypeError(self)

    @ta.final
    def __ne__(self, other):
        raise TypeError(self)


##


@ta.final
class _Cell(BaseCell):
    def __init__(self, v):
        self._v = v

    def __repr__(self):
        return f'_Cell({self._v!r})'

    def __call__(self):
        return self._v

    def set(self, v):
        self._v = v


def cell(v: T) -> Cell[T]:
    return _Cell(v)


#


@ta.final
class _CellCell(BaseCell):
    def __init__(self, c: types.CellType) -> None:
        self._c = c

    def __repr__(self):
        return f'_CellCell({self._c!r})'

    def __call__(self):
        return self._c.cell_contents

    def set(self, v):
        self._c.cell_contents = v


def cell_cell(c: types.CellType) -> Cell:
    return _CellCell(c)



# Copyright (C) 2012-2020 Janne Vanhala
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
# documentation files (the "Software"), to deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all copies or substantial portions of the
# Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
# WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
# https://github.com/jpvanhal/inflection/tree/88eefaacf7d0caaa701af7c8ab2d0ab3f17086f1
"""
A port of Ruby on Rails' inflector to Python.

:copyright: (c) 2012-2020 by Janne Vanhala

:license: MIT, see license above for more details.
"""
import re
import typing as ta
import unicodedata


##


class _RegexReplace(ta.NamedTuple):
    pat: re.Pattern[str]
    rpl: str


_PLURALS: list[_RegexReplace] = [
    _RegexReplace(re.compile(r'(?i)(quiz)$'), r'\1zes'),
    _RegexReplace(re.compile(r'(?i)^(oxen)$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)^(ox)$'), r'\1en'),
    _RegexReplace(re.compile(r'(?i)(m|l)ice$'), r'\1ice'),
    _RegexReplace(re.compile(r'(?i)(m|l)ouse$'), r'\1ice'),
    _RegexReplace(re.compile(r'(?i)(passer)s?by$'), r'\1sby'),
    _RegexReplace(re.compile(r'(?i)(matr|vert|ind)(?:ix|ex)$'), r'\1ices'),
    _RegexReplace(re.compile(r'(?i)(x|ch|ss|sh)$'), r'\1es'),
    _RegexReplace(re.compile(r'(?i)([^aeiouy]|qu)y$'), r'\1ies'),
    _RegexReplace(re.compile(r'(?i)(hive)$'), r'\1s'),
    _RegexReplace(re.compile(r'(?i)([lr])f$'), r'\1ves'),
    _RegexReplace(re.compile(r'(?i)([^f])fe$'), r'\1ves'),
    _RegexReplace(re.compile(r'(?i)sis$'), 'ses'),
    _RegexReplace(re.compile(r'(?i)([ti])a$'), r'\1a'),
    _RegexReplace(re.compile(r'(?i)([ti])um$'), r'\1a'),
    _RegexReplace(re.compile(r'(?i)(buffal|potat|tomat)o$'), r'\1oes'),
    _RegexReplace(re.compile(r'(?i)(bu)s$'), r'\1ses'),
    _RegexReplace(re.compile(r'(?i)(alias|status)$'), r'\1es'),
    _RegexReplace(re.compile(r'(?i)(octop|vir)i$'), r'\1i'),
    _RegexReplace(re.compile(r'(?i)(octop|vir)us$'), r'\1i'),
    _RegexReplace(re.compile(r'(?i)^(ax|test)is$'), r'\1es'),
    _RegexReplace(re.compile(r'(?i)s$'), 's'),
    _RegexReplace(re.compile(r'$'), 's'),
]

_SINGULARS: list[_RegexReplace] = [
    _RegexReplace(re.compile(r'(?i)(database)s$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(quiz)zes$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(matr)ices$'), r'\1ix'),
    _RegexReplace(re.compile(r'(?i)(vert|ind)ices$'), r'\1ex'),
    _RegexReplace(re.compile(r'(?i)(passer)sby$'), r'\1by'),
    _RegexReplace(re.compile(r'(?i)^(ox)en'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(alias|status)(es)?$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(octop|vir)(us|i)$'), r'\1us'),
    _RegexReplace(re.compile(r'(?i)^(a)x[ie]s$'), r'\1xis'),
    _RegexReplace(re.compile(r'(?i)(cris|test)(is|es)$'), r'\1is'),
    _RegexReplace(re.compile(r'(?i)(shoe)s$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(o)es$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(bus)(es)?$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(m|l)ice$'), r'\1ouse'),
    _RegexReplace(re.compile(r'(?i)(x|ch|ss|sh)es$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(m)ovies$'), r'\1ovie'),
    _RegexReplace(re.compile(r'(?i)(s)eries$'), r'\1eries'),
    _RegexReplace(re.compile(r'(?i)([^aeiouy]|qu)ies$'), r'\1y'),
    _RegexReplace(re.compile(r'(?i)([lr])ves$'), r'\1f'),
    _RegexReplace(re.compile(r'(?i)(tive)s$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)(hive)s$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)([^f])ves$'), r'\1fe'),
    _RegexReplace(re.compile(r'(?i)(t)he(sis|ses)$'), r'\1hesis'),
    _RegexReplace(re.compile(r'(?i)(s)ynop(sis|ses)$'), r'\1ynopsis'),
    _RegexReplace(re.compile(r'(?i)(p)rogno(sis|ses)$'), r'\1rognosis'),
    _RegexReplace(re.compile(r'(?i)(p)arenthe(sis|ses)$'), r'\1arenthesis'),
    _RegexReplace(re.compile(r'(?i)(d)iagno(sis|ses)$'), r'\1iagnosis'),
    _RegexReplace(re.compile(r'(?i)(b)a(sis|ses)$'), r'\1asis'),
    _RegexReplace(re.compile(r'(?i)(a)naly(sis|ses)$'), r'\1nalysis'),
    _RegexReplace(re.compile(r'(?i)([ti])a$'), r'\1um'),
    _RegexReplace(re.compile(r'(?i)(n)ews$'), r'\1ews'),
    _RegexReplace(re.compile(r'(?i)(ss)$'), r'\1'),
    _RegexReplace(re.compile(r'(?i)s$'), ''),
]


_UNCOUNTABLE_WORDS: set[str] = {
    'equipment',
    'fish',
    'information',
    'jeans',
    'money',
    'rice',
    'series',
    'sheep',
    'species',
}


def _uncountable_pat(s: str) -> re.Pattern[str]:
    return re.compile(fr'(?i)\b({s})\Z')


_UNCOUNTABLE_PATS: list[re.Pattern[str]] = [
    _uncountable_pat(s) for s in _UNCOUNTABLE_WORDS
]


def _irregular(singular: str, plural: str) -> None:
    """A convenience function to add appropriate rules to plurals and singular for irregular words."""

    def ci(string: str) -> str:
        return ''.join('[' + char + char.upper() + ']' for char in string)

    if singular[0].upper() == plural[0].upper():
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'(?i)({singular[0]}){singular[1:]}$'),
            r'\1' + plural[1:],
        ))
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'(?i)({plural[0]}){plural[1:]}$'),
            r'\1' + plural[1:],
        ))
        _SINGULARS.insert(0, _RegexReplace(
            re.compile(fr'(?i)({plural[0]}){plural[1:]}$'),
            r'\1' + singular[1:],
        ))

    else:
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'{singular[0].upper()}{ci(singular[1:])}$'),
            plural[0].upper() + plural[1:],
        ))
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'{singular[0].lower()}{ci(singular[1:])}$'),
            plural[0].lower() + plural[1:],
        ))
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'{plural[0].upper()}{ci(plural[1:])}$'),
            plural[0].upper() + plural[1:],
        ))
        _PLURALS.insert(0, _RegexReplace(
            re.compile(fr'{plural[0].lower()}{ci(plural[1:])}$'),
            plural[0].lower() + plural[1:],
        ))
        _SINGULARS.insert(0, _RegexReplace(
            re.compile(fr'{plural[0].upper()}{ci(plural[1:])}$'),
            singular[0].upper() + singular[1:],
        ))
        _SINGULARS.insert(0, _RegexReplace(
            re.compile(fr'{plural[0].lower()}{ci(plural[1:])}$'),
            singular[0].lower() + singular[1:],
        ))


_irregular('person', 'people')
_irregular('man', 'men')
_irregular('human', 'humans')
_irregular('child', 'children')
_irregular('sex', 'sexes')
_irregular('move', 'moves')
_irregular('cow', 'kine')
_irregular('zombie', 'zombies')


##


_CAMELIZE_PAT = re.compile(r'(?:^|_)(.)')


def camelize(string: str, uppercase_first_letter: bool = True) -> str:
    """Convert strings to CamelCase."""

    if uppercase_first_letter:
        return _CAMELIZE_PAT.sub(lambda m: m.group(1).upper(), string)
    else:
        return string[0].lower() + camelize(string)[1:]


def dasherize(word: str) -> str:
    """Replace underscores with dashes in the string."""

    return word.replace('_', '-')


_HUMANIZE_PAT_0 = re.compile(r'_id$')
_HUMANIZE_PAT_1 = re.compile(r'(?i)([a-z\d]*)')
_HUMANIZE_PAT_2 = re.compile(r'^\w')


def humanize(word: str) -> str:
    """
    Capitalize the first word and turn underscores into spaces and strip a trailing ``"_id"``, if any. Like
    :func:`titleize`, this is meant for creating pretty output.
    """

    word = _HUMANIZE_PAT_0.sub('', word)
    word = word.replace('_', ' ')
    word = _HUMANIZE_PAT_1.sub(lambda m: m.group(1).lower(), word)
    word = _HUMANIZE_PAT_2.sub(lambda m: m.group(0).upper(), word)
    return word


_ORDINAL_LOOKUP = {
    1: 'st',
    2: 'nd',
    3: 'rd',
}


def ordinal(number: int) -> str:
    """
    Return the suffix that should be added to a number to denote the position in an ordered sequence such as 1st, 2nd,
    3rd, 4th.
    """

    number = abs(int(number))
    if number % 100 in (11, 12, 13):
        return 'th'
    else:
        return _ORDINAL_LOOKUP.get(number % 10, 'th')


def ordinalize(number: int) -> str:
    """
    Turn a number into an ordinal string used to denote the position in an ordered sequence such as 1st, 2nd, 3rd, 4th.
    """

    return f'{number}{ordinal(number)}'


_PARAMETERIZE_PAT = re.compile(r'(?i)[^a-z0-9\-_]+')


def parameterize(string: str, separator: str = '-') -> str:
    """Replace special characters in a string so that it may be used as part of a 'pretty' URL."""

    string = transliterate(string)
    # Turn unwanted chars into the separator
    string = _PARAMETERIZE_PAT.sub(separator, string)
    if separator:
        re_sep = re.escape(separator)
        # No more than one of the separator in a row.
        string = re.sub(fr'{re_sep}{{2,}}', separator, string)
        # Remove leading/trailing separator.
        string = re.sub(fr'(?i)^{re_sep}|{re_sep}$', '', string)

    return string.lower()


def pluralize(word: str) -> str:
    """Return the plural form of a word."""

    if not word or word.lower() in _UNCOUNTABLE_WORDS:
        return word
    else:
        for rule, replacement in _PLURALS:
            if rule.search(word):
                return rule.sub(replacement, word)
        return word


def singularize(word: str) -> str:
    """Return the singular form of a word, the reverse of :func:`pluralize`."""

    for pat in _UNCOUNTABLE_PATS:
        if pat.search(word):
            return word

    for rule, replacement in _SINGULARS:
        if rule.search(word):
            return rule.sub(replacement, word)
    return word


def tableize(word: str) -> str:
    """
    Create the name of a table like Rails does for models to table names. This method uses the :func:`pluralize` method
    on the last word in the string.
    """

    return pluralize(underscore(word))


_TITLEIZE_PAT = re.compile(r"\b('?\w)")


def titleize(word: str) -> str:
    """
    Capitalize all the words and replace some characters in the string to create a nicer looking title. :func:`titleize`
    is meant for creating pretty output.
    """

    return _TITLEIZE_PAT.sub(
        lambda match: match.group(1).capitalize(),
        humanize(underscore(word)).title(),
    )


def transliterate(string: str) -> str:
    """
    Replace non-ASCII characters with an ASCII approximation. If no approximation exists, the non-ASCII character is
    ignored.
    """

    normalized = unicodedata.normalize('NFKD', string)
    return normalized.encode('ascii', 'ignore').decode('ascii')


_UNDERSCORE_PAT_0 = re.compile(r'([A-Z]+)([A-Z][a-z])')
_UNDERSCORE_PAT_1 = re.compile(r'([a-z\d])([A-Z])')


def underscore(word: str) -> str:
    """Make an underscored, lowercase form from the expression in the string."""

    word = _UNDERSCORE_PAT_0.sub(r'\1_\2', word)
    word = _UNDERSCORE_PAT_1.sub(r'\1_\2', word)
    word = word.replace('-', '_')
    return word.lower()



# ruff: noqa: SLF001
import typing as ta

from .. import check
from .. import collections as col
from .. import dataclasses as dc
from .. import lang
from .. import reflect as rfl
from ..text import inflect
from .codecs import Codec
from .fields import Field
from .fields import KeyField
from .fields import RefField
from .indexes import Index
from .keys import Key
from .mappers import Mapper
from .queries import Query
from .refs import Ref
from .registries import Registry
from .sessions import Session
from .sessions import active_session
from .stores import Store


K = ta.TypeVar('K')
T = ta.TypeVar('T')


##


def index(
        *fields: str,
        store_name: str | None = None,
) -> Index:
    check.not_empty(fields)
    for f in fields:
        check.non_empty_str(f)

    return Index(
        _fields=fields,
        _store_name=store_name,
    )


def field(
        name: str,
        ty: ta.Any,
        *,
        store_name: str | None = None,
) -> Field:
    check.non_empty_str(name)

    #

    rty = rfl.type_(ty)

    #

    fld_cls: type[Field] = Field
    if isinstance(rty, rfl.Generic):
        if rty.cls is Key:
            fld_cls = KeyField
        elif rty.cls is Ref:
            fld_cls = RefField
    elif (
            isinstance(rty, rfl.Union) and
            rty.is_optional and
            isinstance(oa := rty.without_none(), rfl.Generic) and
            oa.cls is Ref
    ):
        fld_cls = RefField

    #

    if store_name is None:
        if fld_cls is RefField:
            store_name = name + '_id'
        else:
            store_name = name

    #

    return fld_cls(
        _name=name,
        _store_name=store_name,
        _rty=rty,
    )


def mapper(
        cls: type,
        fields: ta.Sequence[Field] | None = None,
        *,
        store_name: str | None = None,
        indexes: ta.Sequence[Index | str | ta.Sequence[str]] | None = None,
) -> Mapper:
    check.isinstance(cls, type)

    #

    if fields is None:
        if dc.is_dataclass(cls):
            return dataclass_mapper(
                cls,
                store_name=store_name,
                indexes=indexes,
            )
        else:
            raise ValueError(f'Must provide fields for class {cls!r}')

    #

    if store_name is None:
        ns = lang.split_string_casing(cls.__name__)
        store_name = '_'.join([*ns[:-1], inflect.tableize(ns[-1])])

    #

    fields_by_name = col.make_map(((f.name, f) for f in fields), strict=True)

    #

    index_lst: list[Index] = []

    for ia in indexes or ():
        if not isinstance(ia, Index):
            if isinstance(ia, str):
                ia = [ia]
            ia = index(*ia)

        if ia.store_name is None:
            dsn = '__'.join([
                store_name,
                *[fields_by_name[f].store_name for f in ia.fields],
            ])
            ia = ia._with_store_name(dsn)

        index_lst.append(ia)

    #

    return Mapper(
        cls,  # noqa
        store_name,
        fields,
        indexes=index_lst,
    )


def dataclass_mapper(
        cls: type,
        *,
        store_name: str | None = None,
        indexes: ta.Sequence[Index | str | ta.Sequence[str]] | None = None,
) -> Mapper:
    check.arg(dc.is_dataclass(cls))

    dc_rfl = dc.reflect(cls)

    #

    fields: list[Field] = []

    for df in dc_rfl.fields.values():  # noqa
        fields.append(field(df.name, dc_rfl.field_annotations[df.name]))

    #

    return mapper(
        cls,  # noqa
        fields,
        store_name=store_name,
        indexes=indexes,
    )


def registry(
        *mappers: Mapper | type,
        codec: Codec | None = None,
) -> Registry:
    mapper_lst: list[Mapper] = []

    for a in mappers:
        if isinstance(a, Mapper):
            mapper_lst.append(a)
        elif isinstance(a, type):
            mapper_lst.append(mapper(a))
        else:
            raise TypeError(a)

    return Registry(
        mapper_lst,
        codec=codec,
    )


#


def session(
        registry: Registry,  # noqa
        store: Store,
        *,
        no_auto_flush: bool = False,
) -> ta.ContextManager[Session]:
    return Session(
        registry,
        store,
        no_auto_flush=no_auto_flush,
    ).activate()


def abort() -> None:
    active_session().abort()


def add(*objs: ta.Any) -> None:
    return active_session().add(*objs)


@ta.overload
def get(cls: type[T], k: Key[K]) -> T | None:
    ...


@ta.overload
def get(cls: type[T], k: K) -> T | None:
    ...


def get(cls, k):  # noqa
    return active_session().get(cls, k)


def delete(*objs: ta.Any) -> None:
    active_session().delete(*objs)


def flush() -> None:
    active_session().flush()


#


def make_query(
        cls: type[T],
        **where: ta.Any,
) -> Query[T]:
    return Query(
        cls,
        where,
    )


def query(
        cls: type[T],
        **where: ta.Any,
) -> list[T]:
    return active_session().query(make_query(cls, **where))



import typing as ta

from .. import lang


with lang.auto_proxy_import(globals()):
    from . import skiplist
    from .hamt import hamtmap
    from .treap import treapmap


if ta.TYPE_CHECKING:
    from .intersections import PersistentSortedMapping
    from .persistent import PersistentMapping
    from .sorted import SortedMapping
    from .sorted import SortedMutableMapping


K = ta.TypeVar('K')
V = ta.TypeVar('V')


##


def new_persistent_map[K, V](
        items: ta.Iterable[tuple[K, V]] | None = None,
) -> 'PersistentMapping[K, V]':
    if hamtmap.is_hamt_available():
        return hamtmap.new_hamt_map(items)
    else:
        return treapmap.new_treap_map(items)


#


def new_sorted_map[K, V](
        items: ta.Iterable[tuple[K, V]] | None = None,
) -> 'SortedMapping[K, V]':
    return treapmap.new_treap_map(items)


def new_sorted_mutable_map[K, V](
        items: ta.Iterable[tuple[K, V]] | None = None,
) -> 'SortedMutableMapping[K, V]':
    return skiplist.SkipListDict(items)


#


def new_persistent_sorted_map[K, V](
        items: ta.Iterable[tuple[K, V]] | None = None,
) -> 'PersistentSortedMapping[K, V]':
    return treapmap.new_treap_map(items)



import abc
import collections.abc
import operator
import typing as ta
import weakref

from .. import lang


with lang.auto_proxy_import(globals()):
    from . import identity as _identity


K = ta.TypeVar('K')
V = ta.TypeVar('V')

K_co = ta.TypeVar('K_co', covariant=True)
V_co = ta.TypeVar('V_co', covariant=True)


##


def multikey_dict(dct: ta.Mapping[ta.Iterable[K] | K, V], *, deep: bool = False) -> dict[K, V]:
    ret = {}
    for k, v in dct.items():
        if deep and isinstance(v, dict):
            v = multikey_dict(v, deep=True)  # type: ignore
        if isinstance(k, tuple):
            for sk in k:
                ret[sk] = v
        else:
            ret[k] = v
    return ret


##


def guarded_map_update(
        dst: ta.MutableMapping[ta.Any, ta.Any],
        *srcs: ta.Mapping[ta.Any, ta.Any],
) -> ta.MutableMapping[ta.Any, ta.Any]:
    for src in srcs:
        for k, v in src.items():
            if k in dst:
                raise KeyError(k)
            dst[k] = v
    return dst


##


def map_contains(m: ta.Mapping, k: ta.Any) -> bool:
    try:
        m[k]
    except KeyError:
        return False
    else:
        return True


##


class MissingDict(dict[K, V]):
    def __init__(self, missing_fn: ta.Callable[[K], V]) -> None:
        if not callable(missing_fn):
            raise TypeError(missing_fn)

        super().__init__()

        self._missing_fn = missing_fn

    def __missing__(self, key):
        v = self[key] = self._missing_fn(key)
        return v


##


def dict_factory[K, V](
        *,
        identity: bool = False,
        weak: bool = False,
) -> ta.Callable[..., ta.MutableMapping[K, V]]:
    if identity:
        if weak:
            return _identity.IdentityWeakKeyDictionary
        else:
            return _identity.IdentityKeyDict
    elif weak:
        return weakref.WeakKeyDictionary
    else:
        return dict


##


@ta.final
class IterValuesView(collections.abc.ValuesView[V]):
    _mapping: 'IterValuesViewMapping[ta.Any, V]'


class IterValuesViewMapping(collections.abc.Mapping[K, V]):
    @ta.final
    def values(self) -> IterValuesView[V]:
        return IterValuesView(self)

    @abc.abstractmethod
    def itervalues(self) -> ta.Iterator[V]:
        raise NotImplementedError


#


@ta.final
class IterItemsView(collections.abc.ItemsView[K, V]):
    _mapping: 'IterItemsViewMapping[K, V]'

    def __iter__(self) -> ta.Iterator[tuple[K, V]]:
        return self._mapping.iteritems()  # noqa


class IterItemsViewMapping(collections.abc.Mapping[K, V]):
    @ta.final
    def items(self) -> IterItemsView[K, V]:
        return IterItemsView(self)

    @abc.abstractmethod
    def iteritems(self) -> ta.Iterator[tuple[K, V]]:
        raise NotImplementedError


#


class HasIterItems(ta.Protocol[K_co, V_co]):
    def iteritems(self) -> ta.Iterator[tuple[K_co, V_co]]: ...


def iteritems_itervalues(ii: HasIterItems[K, V]) -> ta.Iterator[V]:
    return map(operator.itemgetter(1), ii.iteritems())



import typing as ta

from .. import check
from .keys import _AutoKey
from .mappers import Mapper
from .snaps import Snap
from .stores import Store
from .wrappers import WRAPPER_TYPES


##


class InMemoryStore(Store):
    def __init__(self) -> None:
        super().__init__()

        self._tables_by_mapper: dict[Mapper, InMemoryStore._Table] = {}
        self._tables_by_name: dict[str, InMemoryStore._Table] = {}

    #

    class _Table:
        def __init__(self, m: Mapper) -> None:
            super().__init__()

            self.m = m

            self.snaps: dict[ta.Any, Snap] = {}

            self.indexes: dict[tuple[str, ...], dict[tuple[ta.Any, ...], set[ta.Any]]] = {
                it: {}
                for it in sorted(set(m.index_field_store_names.values()))
            }

            self.index_lookup: dict[frozenset[str], tuple[str, ...]] = {frozenset(t): t for t in self.indexes}

        def __repr__(self) -> str:
            return f'{self.__class__.__name__}({self.m!r})'

    def _table_for_mapper(self, m: Mapper) -> _Table:
        try:
            return self._tables_by_mapper[m]
        except KeyError:
            pass

        check.not_in(m.store_name, self._tables_by_name)

        t = self._Table(m)

        self._tables_by_mapper[m] = t
        self._tables_by_name[m.store_name] = t

        return t

    #

    def fetch(self, m: Mapper, k: ta.Any) -> Snap | None:
        check.not_in(k.__class__, WRAPPER_TYPES)
        try:
            t = self._table_for_mapper(m)
        except KeyError:
            return None
        return t.snaps.get(k)

    #

    def lookup(self, m: Mapper, where: ta.Mapping[str, ta.Any]) -> ta.Sequence[Snap]:
        t = self._table_for_mapper(m)

        if not where:
            return list(t.snaps.values())

        if t.m.key_field.store_name in where:
            if (ks := self.fetch(m, where[t.m.key_field.store_name])) is not None:
                return [ks]
            else:
                return []

        fl = frozenset(where)

        bt: tuple[str, ...] | None = None
        bs: frozenset[str] | None = None
        bi: int | None = None

        for il, it in t.index_lookup.items():
            iz = il - fl
            if iz:
                continue
            di = len(it)
            if not di:
                continue
            if bi is not None and bi > di:
                continue
            bt = it
            bs = il
            bi = di

        if bt is None:
            lst: list[Snap] = []

            for v in t.snaps.values():
                for fk, fv in where.items():
                    if v[fk] != fv:
                        break
                else:
                    lst.append(v)

            return lst

        ik = tuple(where[k] for k in bt)
        ig = [(k, where[k]) for k in fl - bs]  # type: ignore[operator]

        ix = t.indexes[it]  # noqa
        try:
            xs = ix[ik]
        except KeyError:
            return []

        if not ig:
            return [t.snaps[k] for k in xs]

        lst = []

        for k in ig:
            v = t.snaps[k]
            for fk, fv in ig:
                if v[fk] != fv:
                    break
            else:
                lst.append(v)

        return lst

    #

    def flush(self, m: Mapper, b: Store.FlushBatch) -> Store.FlushResult:
        t = self._table_for_mapper(m)

        def index(k: ta.Any, snap: Snap) -> None:  # noqa
            for it, idc in t.indexes.items():
                ik = tuple(snap[f] for f in it)
                try:
                    iz = idc[ik]
                except KeyError:
                    iz = idc[ik] = set()
                iz.add(k)

        def deindex(k: ta.Any, snap: Snap) -> None:  # noqa
            for it, idc in t.indexes.items():
                ik = tuple(snap[f] for f in it)
                iz = idc[ik]
                iz.remove(k)
                if not iz:
                    del idc[ik]

        for snap in b.insert or ():
            k = snap[m.key_field.name]
            kt = k.__class__
            if kt is _AutoKey:
                raise NotImplementedError
            else:
                check.not_in(kt, WRAPPER_TYPES)
            check.not_in(k, t.snaps)
            t.snaps[k] = snap
            index(k, snap)

        for k, snap in b.update or ():
            check.not_in(m.key_field.name, snap)
            kt = k.__class__
            check.not_in(kt, WRAPPER_TYPES)
            x = t.snaps[k]
            deindex(k, x)
            snap = {**x, **snap}
            t.snaps[k] = snap
            index(k, snap)

        for k in b.delete or ():
            snap = t.snaps[k]
            del t.snaps[k]
            deindex(k, snap)

        return self.FlushResult()



import typing as ta

from ... import check
from ... import lang
from ..api.contexts import MarshalContext
from ..api.contexts import UnmarshalContext
from ..api.types import Marshaler
from ..api.types import Unmarshaler
from ..api.values import Value
from ..factories.typemap import TypeMapMarshalerFactory
from ..factories.typemap import TypeMapUnmarshalerFactory


##


class OpaqueReprMarshalerUnmarshaler(Marshaler, Unmarshaler):
    def marshal(self, ctx: MarshalContext, o: ta.Any) -> Value:
        if not isinstance(o, lang.OpaqueRepr):
            o = lang.OpaqueRepr(repr(o))
        return o

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Any:
        return lang.OpaqueRepr(check.isinstance(v, str))


OPAQUE_REPR_MARSHALER_FACTORY = TypeMapMarshalerFactory({lang.OpaqueRepr: OpaqueReprMarshalerUnmarshaler()})
OPAQUE_REPR_UNMARSHALER_FACTORY = TypeMapUnmarshalerFactory({lang.OpaqueRepr: OpaqueReprMarshalerUnmarshaler()})



"""
TODO:
 - MarshalOption, UnmarshalOption, check if a context got something unexpected
 - Options only apply to Factories?
  - MarshalFactoryOption, MarshalOption, UnmarshalFactoryOption, MarshalOption ?
"""
import typing as ta

from ... import collections as col


##


class Option:
    pass


Options: ta.TypeAlias = col.TypeMap[Option]


##


_EMPTY_OPTIONS: Options = col.TypeMap()



import collections.abc
import typing as ta

from ... import check
from ... import dataclasses as dc
from ... import reflect as rfl
from ..api.contexts import MarshalContext
from ..api.contexts import MarshalFactoryContext
from ..api.contexts import UnmarshalContext
from ..api.contexts import UnmarshalFactoryContext
from ..api.options import Options
from ..api.types import Marshaler
from ..api.types import Unmarshaler
from ..api.values import Value
from ..factories.method import MarshalerFactoryMethodClass
from ..factories.method import UnmarshalerFactoryMethodClass
from .api import DefaultMappingConstructors


##


DEFAULT_MAPPING_CONCRETE_TYPES: dict[type[collections.abc.Mapping], type[collections.abc.Mapping]] = {
    collections.abc.Mapping: dict,  # type: ignore
    collections.abc.MutableMapping: dict,  # type: ignore
}


def get_default_mapping_constructor(
        cls: type,
        options: Options | None = None,
) -> ta.Callable[[collections.abc.Mapping], ta.Any]:
    if options is not None and (opt := options.get(DefaultMappingConstructors)) is not None:
        opt = check.isinstance(opt, DefaultMappingConstructors)
        o_ctor: ta.Any = None
        if cls == collections.abc.Mapping:
            o_ctor = opt.mapping
        elif cls == collections.abc.MutableMapping:
            o_ctor = opt.mutable_mapping
        if o_ctor is not None:
            return o_ctor

    return DEFAULT_MAPPING_CONCRETE_TYPES.get(cls, cls)  # noqa


#


@dc.dataclass(frozen=True)
class MappingMarshaler(Marshaler):
    ke: Marshaler
    ve: Marshaler

    def marshal(self, ctx: MarshalContext, o: ta.Mapping) -> Value:
        return {
            self.ke.marshal(ctx, uk): self.ve.marshal(ctx, uv)
            for uk, uv in check.isinstance(o, collections.abc.Mapping).items()
        }


class MappingMarshalerFactory(MarshalerFactoryMethodClass):
    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_generic(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping)):
            return None
        kt, vt = rty.args
        return lambda: MappingMarshaler(ctx.make_marshaler(kt), ctx.make_marshaler(vt))

    @MarshalerFactoryMethodClass.make_marshaler.register
    def _make_concrete(self, ctx: MarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Marshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, collections.abc.Mapping)):
            return None
        return lambda: MappingMarshaler(a := ctx.make_marshaler(ta.Any), a)


#


@dc.dataclass(frozen=True)
class MappingUnmarshaler(Unmarshaler):
    cls: type
    ke: Unmarshaler
    ve: Unmarshaler

    ctor: ta.Callable[[ta.Mapping[ta.Any, ta.Any]], ta.Mapping] | None = None

    def unmarshal(self, ctx: UnmarshalContext, v: Value) -> ta.Mapping:
        dct: dict = {}
        for mk, mv in check.isinstance(v, collections.abc.Mapping).items():
            dct[self.ke.unmarshal(ctx, mk)] = self.ve.unmarshal(ctx, mv)
        if (ctor := self.ctor) is None:
            ctor = get_default_mapping_constructor(self.cls, ctx.options)
        if ctor is dict:
            return dct
        return ctor(dct)


class MappingUnmarshalerFactory(UnmarshalerFactoryMethodClass):
    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_generic(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, rfl.Generic) and issubclass(rty.cls, collections.abc.Mapping)):
            return None
        kt, vt = rty.args
        return lambda: MappingUnmarshaler(rty.cls, ctx.make_unmarshaler(kt), ctx.make_unmarshaler(vt))  # noqa

    @UnmarshalerFactoryMethodClass.make_unmarshaler.register
    def _make_concrete(self, ctx: UnmarshalFactoryContext, rty: rfl.Type) -> ta.Callable[[], Unmarshaler] | None:
        if not (isinstance(rty, type) and issubclass(rty, collections.abc.Mapping)):
            return None
        return lambda: MappingUnmarshaler(check.isinstance(rty, type), a := ctx.make_unmarshaler(ta.Any), a)



# ruff: noqa: SLF001
import typing as ta

from .. import check
from .. import collections as col
from .fields import Field
from .fields import KeyField
from .fields import RefField
from .indexes import Index
from .keys import _KEY_TYPES
from .keys import Key
from .keys import _AutoKey
from .keys import _Key
from .refs import _REF_TYPES
from .refs import _LazyRef
from .snaps import Snap


if ta.TYPE_CHECKING:
    from .registries import Registry


K = ta.TypeVar('K')
T = ta.TypeVar('T')


##


class Mapper(ta.Generic[K, T]):
    def __init__(
            self,
            cls: type[T],
            store_name: str,
            fields: ta.Sequence[Field],
            *,
            indexes: ta.Sequence[Index] | None = None,
    ) -> None:
        super().__init__()

        check.isinstance(cls, type)
        self._cls = cls
        self._store_name = check.non_empty_str(store_name)
        self._fields = fields = check.not_empty(list(fields))
        self._indexes = indexes = list(indexes) if indexes is not None else []

        self._fields_by_name: ta.Mapping[str, Field] = col.make_map(((f.name, f) for f in fields), strict=True)
        self._fields_by_store_name: ta.Mapping[str, Field] = col.make_map(((f.store_name, f) for f in fields), strict=True)  # noqa
        self._store_name_by_field_name = {f.name: f.store_name for f in fields}

        self._key_field: KeyField = check.single(f for f in fields if f.__class__ is KeyField)  # type: ignore[assignment]  # noqa
        self._ref_fields: ta.Sequence[RefField] = [f for f in fields if f.__class__ is RefField]  # type: ignore[misc]

        self._indexes_by_store_name: ta.Mapping[str, Index] = col.make_map((
            (check.non_empty_str(idx.store_name), idx) for idx in indexes
        ), strict=True)
        self._index_field_store_names: ta.Mapping[Index, tuple[str, ...]] = {
            idx: tuple(self._fields_by_name[f].store_name for f in idx.fields)
            for idx in indexes
        }

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._cls!r}, {self._store_name!r})'

    @property
    def cls(self) -> type[T]:
        return self._cls

    @property
    def store_name(self) -> str:
        return self._store_name

    @property
    def fields(self) -> ta.Sequence[Field]:
        return self._fields

    @property
    def indexes(self) -> ta.Sequence[Index]:
        return self._indexes

    #

    @property
    def fields_by_name(self) -> ta.Mapping[str, Field]:
        return self._fields_by_name

    @property
    def fields_by_store_name(self) -> ta.Mapping[str, Field]:
        return self._fields_by_store_name

    @property
    def store_name_by_field_name(self) -> ta.Mapping[str, str]:
        return self._store_name_by_field_name

    @property
    def key_field(self) -> KeyField:
        return self._key_field

    @property
    def ref_fields(self) -> ta.Sequence[RefField]:
        return self._ref_fields

    @property
    def indexes_by_store_name(self) -> ta.Mapping[str, Index]:
        return self._indexes_by_store_name

    @property
    def index_field_store_names(self) -> ta.Mapping[Index, tuple[str, ...]]:
        return self._index_field_store_names

    #

    _registry: 'Registry'

    def _set_registry(self, r: 'Registry') -> None:
        try:
            self._registry  # noqa
        except AttributeError:
            pass
        else:
            raise RuntimeError('registry already set')
        self._registry = r

    @property
    def registry(self) -> 'Registry':
        return self._registry

    #

    def key_for_obj(self, obj: T) -> Key[K]:
        return check.isinstance(getattr(obj, self.key_field.name), Key)

    def key_for_snap(self, snap: Snap) -> Key[K]:
        k = snap[self._key_field._name]
        kt = k.__class__
        if kt is _AutoKey:
            return k
        check.not_in(kt, _KEY_TYPES)
        return _Key(k)

    #

    def field_value_to_snap_value(self, f: Field, v: ta.Any) -> ta.Any:
        vt = v.__class__
        vk = vt in _KEY_TYPES
        vr = vt in _REF_TYPES

        ft = f.__class__
        if ft is KeyField:
            check.state(vk)
            if vt is _AutoKey:
                raise NotImplementedError
            else:
                v = v.k

        elif ft is RefField:
            if f._optional and v is None:  # type: ignore[attr-defined]
                pass
            else:
                check.state(vr)
                v = v.k
                if vt is _AutoKey:
                    raise NotImplementedError
                else:
                    v = v.k

        else:
            check.state(not vk)
            check.state(not vr)

            if (co := self._registry._codec) is not None:
                v = co.decode(v, f._rty)

        return v

    def obj_to_snap(self, obj: T) -> Snap:
        dct: dict[str, ta.Any] = {}

        for f in self._fields:
            dct[f._store_name] = self.field_value_to_snap_value(f, getattr(obj, f.name))

        return dct

    #

    def snap_value_to_field_value(self, f: Field, v: ta.Any) -> ta.Any:
        ft = f.__class__
        if ft is KeyField:
            v = _Key(v)

        elif ft is RefField:
            if v is None:
                check.state(f._optional)  # type: ignore[attr-defined]
            else:
                v = _LazyRef(f._ref_obj_cls, _Key(v))  # type: ignore[attr-defined]

        elif (co := self._registry._codec) is not None:
            v = co.decode(v, f._rty)

        return v

    def snap_to_obj(self, snap: Snap) -> T:
        check.equal(len(snap), len(self._fields))

        kw: dict[str, ta.Any] = {}

        for f in self._fields:
            kw[f._name] = self.snap_value_to_field_value(f, snap[f._store_name])

        return self._cls(**kw)



import typing as ta

from .. import lang


T = ta.TypeVar('T')


##


@ta.final
class Query(lang.Final, ta.Generic[T]):
    def __init__(
            self,
            cls: type[T],
            where: ta.Mapping[str, ta.Any],
    ) -> None:
        super().__init__()

        self._cls = cls
        self._where = where

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}({self._cls!r}, {self._where!r})'

    @property
    def cls(self) -> type[T]:
        return self._cls

    @property
    def where(self) -> ta.Mapping[str, ta.Any]:
        return self._where



import typing as ta

from .. import check
from .. import lang


##


@ta.final
class Index(lang.Final):
    def __init__(
            self,
            *,
            _fields: ta.Sequence[str],
            _store_name: str | None = None,
    ) -> None:
        super().__init__()

        self._fields = tuple(check.non_empty_str(f) for f in check.not_empty(_fields))
        self._store_name = check.non_empty_str(_store_name) if _store_name is not None else None

        check.unique(self._fields)

    def _with_store_name(self, store_name: str) -> 'Index':
        return Index(
            _fields=self._fields,
            _store_name=store_name,
        )

    def __repr__(self) -> str:
        return (
            f'{type(self).__name__}('
            f'{list(self._fields)!r}'
            f"{f', store_name={self._store_name!r})' if self._store_name else ''}"
            f')'
        )

    @property
    def fields(self) -> tuple[str, ...]:
        return self._fields

    @property
    def store_name(self) -> str | None:
        return self._store_name



import abc
import typing as ta

from .. import lang
from .. import marshal as msh
from .. import reflect as rfl


##


class Codec(lang.Abstract):
    @abc.abstractmethod
    def encode(self, obj: ta.Any, rty: rfl.Type) -> ta.Any:
        raise NotImplementedError

    @abc.abstractmethod
    def decode(self, val: ta.Any, rty: rfl.Type) -> ta.Any:
        raise NotImplementedError


@ta.final
class NopCodec(Codec):
    def encode(self, obj: ta.Any, rty: rfl.Type) -> ta.Any:
        return obj

    def decode(self, val: ta.Any, rty: rfl.Type) -> ta.Any:
        return val


@ta.final
class FnCodec(Codec):
    def __init__(
            self,
            encode: ta.Callable[[ta.Any, rfl.Type], ta.Any],
            decode: ta.Callable[[ta.Any, rfl.Type], ta.Any],
    ) -> None:
        super().__init__()

        self._encode = encode
        self._decode = decode

    def encode(self, obj: ta.Any, rty: rfl.Type) -> ta.Any:
        return self._encode(obj, rty)

    def decode(self, val: ta.Any, rty: rfl.Type) -> ta.Any:
        return self._decode(val, rty)


##


@ta.final
class MarshalCodec(Codec):
    def encode(self, obj: ta.Any, rty: rfl.Type) -> ta.Any:
        return msh.marshal(obj, rty)

    def decode(self, val: ta.Any, rty: rfl.Type) -> ta.Any:
        return msh.unmarshal(val, rty)
